from typing import Dict, List, Optional
import json
import pathlib
import webbrowser
import sys
import os
import argparse

from antlr4 import * # type: ignore

import harmony_model_checker
from harmony_model_checker.config import settings
from harmony_model_checker import charm # type: ignore
from harmony_model_checker.exception import HarmonyCompilerError, HarmonyCompilerErrorCollection
import harmony_model_checker.harmony.harmony as legacy_harmony
from harmony_model_checker.harmony.genhtml import GenHTML
from harmony_model_checker.harmony.brief import Brief
from harmony_model_checker.harmony.verbose import Verbose
from harmony_model_checker.compile import do_compile


args = argparse.ArgumentParser(
    "harmony", description="Harmony programming language compiler and model checker")
args.add_argument("-a", action="store_true",
                  help="list machine code (with labels)")
args.add_argument("-b", action="store_true",
                  help="enable behavior subset check")
args.add_argument("-T", action="store_true",
                  help="print timing info")
args.add_argument("-A", action="store_true",
                  help="list machine code (without labels)")
args.add_argument("-B", type=str, help="check against the given behavior")
args.add_argument("-d", "--direct", action="store_true",
                  help="run directly without model checking")
args.add_argument("-p", "--parse", action="store_true",
                  help="parse code without running")
args.add_argument("-c", "--const", action='append', type=str,
                  metavar="name=value", help="define a constant")
args.add_argument("-D", action="store_true",
                  help="dump Kripke graph")
args.add_argument("-W", action="store_true",
                  help="suppress busy waiting check")
args.add_argument("-R", action="store_true",
                  help="suppress race condition warnings")
args.add_argument("-U", action="store_true",
                  help="do not pin workers")
args.add_argument("--module", "-m", action="append", type=str,
                  metavar="module=version", help="select a module version")
args.add_argument("-i", "--intf", type=str, metavar="expr",
                  help="specify an interface function")
args.add_argument("-s", action="store_true",
                  help="silent (do not print periodic status updates)")
args.add_argument("-v", "--version", action="store_true",
                  help="print version number")
args.add_argument("--quick", action="store_true",
                  help="no post-model-check analysis")
args.add_argument("-o", action='append', type=pathlib.Path,
                  help="specify output file (.hvm, .hco, .hfa, .htm. .tla, .tex, .png, .gv)")
args.add_argument("-j", action="store_true",
                  help="list machine code in JSON format")
args.add_argument("-w", type=str, help="set number of workers")
args.add_argument("--noweb", action="store_true", default=False,
                  help="do not automatically open web browser")
args.add_argument("--suppress", action="store_true",
                  help="generate less terminal output")
args.add_argument("--config", action="store_true",
                  help="get or set configuration value. "
                       "Use --config <key> to get the value of a setting. "
                       "Use --config <key> <value> to set the value of a setting")

# Internal flags
args.add_argument("--cf", action="append", type=str, help=argparse.SUPPRESS)
args.add_argument("args", metavar="args", type=str, nargs='*', help="arguments")

def handle_hny(ns, output_files, parse_code_only, filenames):
    for suffix, file in output_files.items():
        if file is not None:
            try:
                os.remove(file)
            except:
                pass

    print("* Phase 1: compile Harmony program to bytecode", flush=True)

    consts: List[str] = ns.const or []
    interface: Optional[str] = ns.intf
    mods: List[str] = ns.module or []

    try:
        code, scope = do_compile(filenames, consts, mods, interface)
    except (HarmonyCompilerErrorCollection, HarmonyCompilerError) as err:
        if isinstance(err, HarmonyCompilerErrorCollection):
            errors = err.errors
        else:
            errors = [err.token]

        if parse_code_only:
            data = dict(errors=[e._asdict() for e in errors], status="error")
            with open(output_files["hvm"], "w", encoding='utf-8') as fp:
                json.dump(data, fp)
        else:
            for e in errors:
                print(f"Line {e.line}:{e.column} at {e.filename}, {e.message}")
                print()
        exit(1)

    if parse_code_only:
        with open(output_files["hvm"], "w", encoding='utf-8') as f:
            f.write(json.dumps({"status": "ok"}))
        exit()

    if output_files["tla"] is not None:
        with open(output_files["tla"], "w", encoding='utf-8') as f:
            legacy_harmony.tla_translate(f, code, scope)

    if output_files["tex"] is not None:
        with open(output_files["tex"], "w", encoding='utf-8') as f:
            legacy_harmony.tex_output(f, code, scope)
    
    return code, scope

def handle_hvm(ns, output_files, parse_code_only, code, scope):
    charm_options = ns.cf or []
    if ns.B:
        charm_options.append("-B" + ns.B)
    if ns.b:
        charm_options.append("-b")
    if ns.T:
        charm_options.append("-T")
    if ns.w:
        charm_options.append("-w" + ns.w)
    if ns.D:
        charm_options.append("-D")
    if ns.R:
        charm_options.append("-R")
    if ns.U:
        charm_options.append("-U")
    if ns.W:
        charm_options.append("-c")

    # see if there is a configuration file
    if code is not None:
        with open(output_files["hvm"], "w", encoding='utf-8') as fd:
            legacy_harmony.dumpCode("json", code, scope, f=fd)

    if parse_code_only:
        exit()

    if ns.direct:
        # print("* Phase 2: run", flush=True)
        charm_options.append("-d")
        r = charm.run_model_checker(
            *charm_options,
            output_files["hvm"]
        )
        if r != 0:
            print("charm failed")
            exit(r)
    else:
        # print("* Phase 2: run the model checker", flush=True)
        if "hfa" in output_files and output_files["hfa"] != None:
            r = charm.run_model_checker(
                *charm_options,
                "-o" + output_files["hco"],
                "-o" + output_files["hfa"],
                output_files["hvm"]
            )
        else:
            r = charm.run_model_checker(
                *charm_options,
                "-o" + output_files["hco"],
                output_files["hvm"]
            )
        if r != 0:
            print("charm model checker failed")
            exit(r)

def handle_hco(ns, output_files):
    if ns.quick:
        return

    suppress_output = ns.suppress

    behavior = None
    if ns.B:
        behavior = ns.B

    disable_browser = settings.values.disable_web or ns.noweb
    
    b = Brief()
    b.run(output_files, behavior)
    vb = Verbose()
    vb.run(output_files)
    gh = GenHTML()
    gh.run(output_files)
    if not suppress_output:
        print()
        p = pathlib.Path(output_files["htm"]).resolve()
        url = "file://" + str(p)
        print("open " + url + " for detailed information, or use the HarmonyGUI", file=sys.stderr)
        if not disable_browser:
            webbrowser.open(url)
    # exit as minify may still be running
    print(flush=True, end="")
    print(file=sys.stderr, flush=True, end="")
    os._exit(0)
    
def handle_version(_: argparse.Namespace):
    print("Version:", harmony_model_checker.__package__,
          harmony_model_checker.__version__)
    return 0


def handle_config(ns: argparse.Namespace):
    if len(ns.args) == 0:
        print("Configuration settings:")
        for k, v in settings.values._asdict().items():
            print(f"    {k}: {v}")
        print("Use --config <key> to get the value of a setting.\nUse --config <key> <value> to set the value of a setting")
        return 0
    key = ns.args[0]
    try:
        if len(ns.args) > 1:
            value = ns.args[1]
            settings.update_settings_file(key, value)
        else:
            print(settings.get_settings_value(key))
    except AttributeError:
        print(f"'{key}' is not a valid configuration setting")
        return 1
    except ValueError as e:
        print(
            f"Value '{e.args[0]}' is not a valid value for configuration setting '{key}'")
        return 1
    return 0


def main():
    ns = args.parse_args()

    if ns.version:
        return handle_version(ns)

    if ns.config:
        return handle_config(ns)

    parse_code_only: bool = ns.parse
    legacy_harmony.silent = ns.s
    
    output_files: Dict[str, Optional[str]] = {
        "hfa": None,
        "htm": None,
        "hvb": None,
        "hco": None,
        "hvm": None,
        "png": None,
        "tla": None,
        "tex": None,
        "gv":  None
    }
    for p in (ns.o or []):
        # The suffix includes the dot if it exists.
        # Otherwise, it is an empty string.
        suffix = p.suffix[1:]
        if suffix not in output_files:
            print(f"Unknown file suffix on {p}")
            return 1
        if output_files[suffix] is not None:
            print(f"Duplicate suffix '.{suffix}'")
            return 1
        output_files[suffix] = str(p)

    if len(ns.args) != 1:
        print(f"harmony: error: invalid number of arguments ({len(ns.args)}). Provide 1 argument.")
        args.print_help()
        return 1

    filename = pathlib.Path(ns.args[0])
    if not filename.exists():
        print(f"harmony: error: file named '{filename}' does not exist.")
        return 1

    stem = str(filename.parent / filename.stem)
    input_file_type = filename.suffix

    if output_files["hvm"] is None:
        output_files["hvm"] = stem + ".hvm"
    if output_files["hco"] is None:
        output_files["hco"] = stem + ".hco"
    if output_files["htm"] is None:
        output_files["htm"] = stem + ".htm"
    if output_files["hvb"] is None:
        output_files["hvb"] = stem + ".hvb"
    if output_files["hfa"] is None:
        output_files["hfa"] = stem + ".hfa"
    if output_files["png"] is None:
        output_files["png"] = stem + ".png"
    if output_files["png"] is not None and output_files["gv"] is None:
        output_files["gv"] = stem + ".gv"

    charm_flag = True
    print_code: Optional[str] = None
    if ns.a:
        print_code = "verbose"
        charm_flag = False
    if ns.A:
        print_code = "terse"
        charm_flag = False
    if ns.j:
        print_code = "json"
        charm_flag = False

    # Handle different Harmony compilation stages
    if input_file_type == ".hny":
        code, scope = handle_hny(ns, output_files, parse_code_only, str(filename))
        if charm_flag:
            handle_hvm(ns, output_files, parse_code_only, code, scope)
            handle_hco(ns, output_files)
        else:
            print("Skipping Phases 2-5...", flush=True)
            legacy_harmony.dumpCode(print_code, code, scope)

    if input_file_type == ".hvm":
        print("Skipping Phase 1...", flush=True)
        handle_hvm(ns, output_files, parse_code_only, None, None)
        handle_hco(ns, output_files)
        
    if input_file_type == ".hco":
        print("Skipping Phases 1-4...", flush=True)
        handle_hco(ns, output_files)



from typing import Optional
from harmony_model_checker.exception import *

class Scope:
    def __init__(self, parent: Optional['Scope']):
        self.parent = parent               # parent scope
        self.names = { "this": ("local-var", ("this", "NOFILE", 0, 0)) }   # name to (type, x) map
        self.labels: dict = {} if parent is None else parent.labels
        self.prefix: Optional[str] = None if parent is None else parent.prefix
        self.inherit = False
        self.pmap: dict = {}                  # hack for pretty-printing
        self.file = "__nofile__"              # file name
        self.uses_pre = False                 # for invariant optimization

    def copy(self):
        c = Scope(self.parent)
        c.names = self.names.copy()
        c.labels = self.labels.copy()
        c.prefix = self.prefix
        c.inherit = self.inherit
        return c

    def checkUnused(self, name):
        # tv = self.names.get(lexeme)
        (lexeme, file, line, column) = name
        if lexeme != '_' and lexeme != 'result' and self.find(name):
            raise HarmonyCompilerError(
                lexeme=lexeme,
                filename=file,
                # stmt=stmt,
                line=line,
                column=column,
                message='variable %s shadows previous use' % lexeme
            )
            # (t, v) = tv
            # assert t != "variable", ("variable name in use", name, v)

    def lookup(self, name):
        (lexeme, file, line, column) = name
        if lexeme == "_":
            return ("local-var", name)
        tv = self.names.get(lexeme)
        if tv is not None:
            return tv
        ancestor = self.parent
        while ancestor is not None:
            tv = ancestor.names.get(lexeme)
            if tv is not None:
                # (t, v) = tv
                # if t == "local-var":
                #    return None
                return tv
            ancestor = ancestor.parent
        # print("Warning: unknown name:", name, " (assuming global variable)")  TODO
        if lexeme != '_':
            self.names[lexeme] = ("global", name)
        return ("global", name)

    # like lookup but returns error if not found
    def find(self, name):
        (lexeme, file, line, column) = name
        if lexeme == "_":
            return ("local-var", name)
        tv = self.names.get(lexeme)
        if tv is not None:
            return tv
        ancestor = self.parent
        while ancestor is not None:
            tv = ancestor.names.get(lexeme)
            if tv is not None:
                # (t, v) = tv
                # if t == "local-var":
                #    return None
                return tv
            ancestor = ancestor.parent

        return False

    # This is a hack to create a map of identifiers to types for
    # pretty printing.
    # TODO: come up with a better solution
    def pset(self, lexeme: str, tv):
        if lexeme == '_':
            return
        self.pmap[lexeme] = tv
        if self.parent is not None:
            self.parent.pset(lexeme, tv)

    def set(self, lexeme, tv):
        if lexeme == '_':
            return
        self.names[lexeme] = tv
        self.pset(lexeme, tv)

    def tryset(self, token, tv):
        (lexeme, file, line, column) = token
        if lexeme == '_':
            return
        if lexeme in self.names:
            raise HarmonyCompilerError(
                filename=file,
                lexeme=lexeme,
                line=line,
                column=column,
                message="%s: error: duplicate defined in this module" % lexeme
            )
        self.names[lexeme] = tv
        self.pset(lexeme, tv)



import unittest


from harmony_model_checker.h2py.h2py_runtime import H, P, HAddr, HDict, hcompare


class TestH2PyRuntimeH(unittest.TestCase):

    def test_none(self):
        self.assertIsNone(H(None))

    def test_int(self):
        self.assertEqual(H(42), 42)

    def test_true(self):
        self.assertEqual(H(True), True)

    def test_false(self):
        self.assertEqual(H(False), False)

    def test_dict_1(self):
        self.assertEqual(
            H({True: False, 0: 1}),
            HDict({True: False, 0: 1})
        )

    def test_dict_2(self):
        self.assertEqual(
            H({H({True: False}): H({0: 1})}),
            HDict({HDict({True: False}): HDict({0: 1})})
        )

    def test_addr_1(self):
        self.assertIsNone(H(None))


class TestH2PyRuntimeP(unittest.TestCase):

    def test_none(self):
        self.assertIsNone(P(None))

    def test_int(self):
        self.assertEqual(P(42), 42)

    def test_true(self):
        self.assertEqual(P(True), True)

    def test_false(self):
        self.assertEqual(P(False), False)


class TestH2PyRuntimeHcompare(unittest.TestCase):

    def test_typeorder(self):
        self.assertLess(hcompare(H(False), H(42)), 0)
        self.assertLess(hcompare(H(False), H("abc")), 0)
        self.assertLess(hcompare(H(False), H({})), 0)
        self.assertLess(hcompare(H(False), HAddr(("x",))), 0)
        self.assertLess(hcompare(H(42), H("abc")), 0)
        self.assertLess(hcompare(H(42), H({})), 0)
        self.assertLess(hcompare(H(42), HAddr(("x",))), 0)
        self.assertLess(hcompare(H("abc"), H({})), 0)
        self.assertLess(hcompare(H("abc"), HAddr(("x",))), 0)
        self.assertLess(hcompare(H({}), HAddr(("x",))), 0)

    def test_int(self):
        self.assertLess(hcompare(H(42), H(43)), 0)
        self.assertEqual(hcompare(H(42), H(42)), 0)
        self.assertGreater(hcompare(H(43), H(42)), 0)

    def test_bool(self):
        self.assertEqual(hcompare(H(False), H(False)), 0)
        self.assertLess(hcompare(H(False), H(True)), 0)
        self.assertGreater(hcompare(H(True), H(False)), 0)
        self.assertEqual(hcompare(H(True), H(True)), 0)

    def test_atom(self):
        self.assertLess(hcompare(H("abc"), H("abcd")), 0)
        self.assertLess(hcompare(H("abc"), H("abd")), 0)
        self.assertEqual(hcompare(H("abc"), H("abc")), 0)
        self.assertGreater(hcompare(H("abcd"), H("abc")), 0)
        self.assertGreater(hcompare(H("abd"), H("abc")), 0)

    def test_dict(self):
        self.assertLess(hcompare(H({}), H({0: 1})), 0)
        self.assertEqual(hcompare(H({0: 1}), H({0: 1})), 0)
        self.assertGreater(hcompare(H({0: 1, 2: 3}), H({0: 1})), 0)

    def test_addr(self):
        self.assertLess(hcompare(H(None), HAddr(("x",))), 0)
        self.assertEqual(hcompare(HAddr(("x",)), HAddr(("x",))), 0)
        self.assertGreater(hcompare(HAddr(("x", "y")), HAddr(("x",))), 0)


if __name__ == '__main__':
    unittest.main()




import json
import os
import pathlib
from typing import NamedTuple


class Setting(NamedTuple):
    disable_web: bool

def _make_new_settings():
    return Setting(
        disable_web=False,
    )
_default_settings = _make_new_settings()


class SettingsManager:
    def __init__(self) -> None:
        self.user_dir = self._setup_harmony_dir()
        self.settings_file = self._setup_settings_file()
        
        self.values = self._load_settings_file()
        self._update_from_env()

    def update_settings_file(self, attr, value):
        value_type = type(_default_settings.__getattribute__(attr))        

        with self.settings_file.open('r') as f:
            d = json.load(f)
        
        if value_type == bool:
            v = str(value).lower() == 'true'
        else:
            raise ValueError(value)

        d[attr] = v
        with self.settings_file.open('w') as f:
            json.dump(d, f)

    def get_settings_value(self, attr):
        return self.values.__getattribute__(attr)

    def _setup_harmony_dir(self):
        """Create a application-specific directory following the XDG specification."""
        xdg_config_home = os.environ.get('XDG_CONFIG_HOME', str(pathlib.Path.home() / ".config"))
        user_dir = pathlib.Path(xdg_config_home, 'harmony-model-checker').expanduser()
        if not user_dir.is_dir():
            user_dir.mkdir(parents=True)
        return user_dir

    def _setup_settings_file(self):
        settings_path = self.user_dir.joinpath('settings.json')
        if not settings_path.is_file():
            s = _make_new_settings()
            with settings_path.open('w') as f:
                json.dump(s._asdict(), f)
        return settings_path

    def _load_settings_file(self):
        with self.settings_file.open('r') as f:
            try:
                d: dict = json.load(f)
                v = d.get('disable_web')
                if v is not None and isinstance(v, bool):
                    return Setting(disable_web=v)
            except TypeError:
                pass
            return _make_new_settings()

    def _update_from_env(self):
        val = os.environ.get('HARMONY_DISABLE_WEB', None)
        if val is not None:
            self.values.disable_web = val.lower() == 'true'


settings = SettingsManager()



from harmony_model_checker.harmony.ast import *


class AbstractASTVisitor:

    def _not_implemented(self, node, *args, **kwargs):
        raise NotImplementedError(f'Not implemented: {type(node)}')

    def visit_constant(self, node: ConstantAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_name(self, node: NameAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_set(self, node: SetAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_range(self, node: RangeAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_tuple(self, node: TupleAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_dict(self, node: DictAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_set_comprehension(self, node: SetComprehensionAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_dict_comprehension(self, node: DictComprehensionAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_list_comprehension(self, node: ListComprehensionAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_nary(self, node: NaryAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_cmp(self, node: CmpAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_apply(self, node: ApplyAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_pointer(self, node: PointerAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_assignment(self, node: AssignmentAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_del(self, node: DelAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_set_int_level(self, node: SetIntLevelAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_save(self, node: SaveAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_stop(self, node: StopAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_address(self, node: AddressAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_pass(self, node: PassAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_block(self, node: BlockAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_if(self, node: IfAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_while(self, node: WhileAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_invariant(self, node: InvariantAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_let(self, node: LetAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_var(self, node: VarAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_for(self, node: ForAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_let_when(self, node: LetWhenAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_atomic(self, node: AtomicAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_assert(self, node: AssertAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_print(self, node: PrintAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_method(self, node: MethodAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_lambda(self, node: LambdaAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_call(self, node: CallAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_spawn(self, node: SpawnAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_trap(self, node: TrapAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_go(self, node: GoAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_import(self, node: ImportAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_from(self, node: FromAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_location(self, node: LocationAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_label_stat(self, node: LabelStatAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_sequential(self, node: SequentialAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)

    def visit_const(self, node: ConstAST, *args, **kwargs):
        self._not_implemented(node, *args, **kwargs)



def json_idx(js):
    if js["type"] == "atom":
        return json_string(js)
    return "[" + json_string(js) + "]"

def json_string(js):
    type = js["type"]
    if type == "address":
        if "func" not in js:
            return "None"
        result = "?"
        func = js["func"]
        args = js["args"]
        if func["type"] == "pc":
            if int(func["value"]) == -1:
                result += args[0]["value"]
                args = args[1:]
            elif int(func["value"]) == -2:
                result += "@" + args[0]["value"]
                args = args[1:]
            else:
                result += json_string(func)
        else:
            result += json_string(func)
        return result + "".join([ json_idx(kv) for kv in args ])
    v = js["value"]
    if type in { "bool", "int" }:
        return v
    if type == "atom":
        return '"' + v + '"'
    if type == "list":
        if v == []:
            return "[]"
        return "[ " + ", ".join([ json_string(val) for val in v]) + " ]"
    if type == "set":
        if v == []:
            return "{}"
        return "{ " + ", ".join([ json_string(val) for val in v]) + " }"
    if type == "dict":
        if v == []:
            return "{:}"
        lst = [ (json_string(js["key"]), json_string(js["value"])) for js in v ]
        # if [ k for k,_ in lst ] == [ str(i) for i in range(len(v)) ]:
        #     return "[ " + ", ".join([ x for _,x in lst ]) + " ]" 
        return "{ " + ", ".join([ k + ": " + x for k,x in lst ]) + " }" 
    if type == "pc":
        return "PC(%s)"%v
    if type == "address":
        if v == []:
            return "None"
        return "?" + v[0]["value"] + "".join([ json_idx(kv) for kv in v[1:] ])
    if type == "context":
        return "CONTEXT(" + str(v["pc"]) + ")"
    assert False



from antlr4 import InputStream, FileStream, CommonTokenStream # type: ignore

from harmony_model_checker.exception import HarmonyCompilerError, HarmonyCompilerErrorCollection
from harmony_model_checker.harmony.ast import AST, BlockAST

import harmony_model_checker.harmony.harmony as legacy_harmony
from harmony_model_checker.harmony.harmony import Scope, Code, State
from harmony_model_checker.harmony.ops import FrameOp, ReturnOp
from harmony_model_checker.harmony.value import AddressValue, ContextValue
from harmony_model_checker.parser.antlr_rule_visitor import HarmonyVisitorImpl
from harmony_model_checker.parser.HarmonyParser import HarmonyParser
from harmony_model_checker.parser.HarmonyErrorListener import HarmonyLexerErrorListener, HarmonyParserErrorListener
from harmony_model_checker.parser.HarmonyLexer import HarmonyLexer

import os

from typing import List, Optional, Tuple

def _build_input_stream(**kwargs) -> InputStream:
    try:
        filename = kwargs.get('filename', None)
        str_value = kwargs.get('str_value', None)
        if filename is not None:
            return FileStream(filename, 'utf-8')
        elif str_value is not None:
            return InputStream(str_value)
    except UnicodeDecodeError as e:
        lexeme = str(e.args[1][e.start:e.end]) # e.args[0] is the encoding name, and e.args[1] contains the bytes being parsed.
        raise HarmonyCompilerError(
            message=e.reason,
            filename=filename,
            line=1,
            column=1,
            lexeme=lexeme
        )
    raise ValueError("Cannot build input stream without a source")


def _build_parser(progam_input: InputStream, lexer_error_listener=None, parser_error_listener=None):
    lexer = HarmonyLexer(progam_input)
    stream = CommonTokenStream(lexer)
    parser = HarmonyParser(stream)

    lexer.removeErrorListeners()
    parser.removeErrorListeners()
    if lexer_error_listener:
        lexer.addErrorListener(lexer_error_listener)
    if parser_error_listener:
        parser.addErrorListener(parser_error_listener)

    return parser


def _load_string(string, scope: Scope, code: Code, filename="<string-code>"):
    legacy_harmony.namestack.append(filename)
    ast = _parse_string(string, filename)
    for mod in ast.getImports():
        _do_import(scope, code, mod)

    # method names and label names get a temporary value
    # they are filled in with their actual values after compilation
    # TODO.  Look for duplicates?
    for (token, lb) in ast.getLabels():
        (lexeme, file, line, column) = token
        if scope.find(token):
            raise HarmonyCompilerError(
                message="Duplicate label %s (defined before in line %d)" % (lexeme, scope.names[lexeme][1][2]),
                filename=file,
                line=line,
                column=column,
                lexeme=lexeme
            )
        scope.names[lexeme] = ("constant", (lb, file, line, column))

    ast.compile(scope, code, None)
    legacy_harmony.namestack.pop()


def _load_file(filename: str, scope: Scope, code: Code, init: bool):
    if filename in legacy_harmony.files:
        return
    legacy_harmony.namestack.append(filename)
    with open(filename, "r", encoding='utf-8') as f:
        legacy_harmony.files[filename] = f.read().split("\n")

    ast = _parse(filename)
    if ast is None:
        raise HarmonyCompilerError(
            message="Unknown error: unable to parse Harmony file",
            filename=filename
        )

    _, _, line1, column1 = ast.token
    lexeme2, _, line2, column2 = ast.endtoken
    column2 += len(lexeme2) - 1
    if init:
        code.append(FrameOp(("__init__", None, None, None), []), ast.token, ast.endtoken, stmt=(line1, column1, line2, column2))

    for mod in ast.getImports():
        _do_import(scope, code, mod)

    # method names and label names get a temporary value
    # they are filled in with their actual values after compilation
    # TODO.  Look for duplicates?
    for (token, lb) in ast.getLabels():
        (lexeme, file, line, column) = token
        if scope.find(token):
            raise HarmonyCompilerError(
                message="Duplicate label %s" % lexeme,
                filename=file,
                line=line,
                column=column,
                lexeme=lexeme
            )
        scope.names[lexeme] = ("constant", (lb, file, line, column))

    ast.compile(scope, code, None)
    if init:
        _, file, line, column = ast.token
        code.append(ReturnOp(("result", file, line, column), AddressValue(None, [])), ast.token, ast.endtoken, stmt=(line1, column1, line2, column2))  # to terminate "__init__" process
    legacy_harmony.namestack.pop()


def _do_import(scope: Scope, code: Code, module: Tuple[str, Optional[str], Optional[int], Optional[int]]):
    (lexeme, file, line, column) = module
    code.modpush(lexeme)
    # assert lexeme not in scope.names        # TODO
    if lexeme not in legacy_harmony.imported:
        # module name replacement with -m flag
        modname = legacy_harmony.modules.get(lexeme, lexeme)

        # create a new scope
        scope2 = Scope(None)
        scope2.prefix = lexeme
        scope2.labels = scope.labels
        scope2.inherit = True

        found = False
        install_path = os.path.dirname(os.path.realpath(__file__))
        for directory in [os.path.dirname(legacy_harmony.namestack[-1]), os.path.join(install_path, "modules"), "."]:
            filename = os.path.join(directory, modname + ".hny")
            if os.path.exists(filename):
                scope2.file = filename
                _load_file(filename, scope2, code, False)
                found = True
                break
        if not found:
            raise HarmonyCompilerError(
                filename=file,
                lexeme=modname,
                message="Can't import module %s from %s" % (
                    modname, legacy_harmony.namestack),
                line=line,
                column=column
            )
        legacy_harmony.imported[lexeme] = scope2

    scope.names[lexeme] = ("module", legacy_harmony.imported[lexeme])
    code.modpop()

def _parse_constant(name: str, value: str):
    filename = "<constant argument>"
    _input = _build_input_stream(str_value=value)
    parser = _build_parser(_input)
    visitor = HarmonyVisitorImpl(filename)

    tree = parser.expr()
    ast = visitor.visit(tree)

    scope = Scope(None)
    code = Code()
    code.modpush("__const__")
    ast.compile(scope, code, None)
    state = State(code, scope.labels)
    ctx = ContextValue(("__arg__", None, None, None), 0,
                       legacy_harmony.emptytuple, legacy_harmony.emptytuple)
    ctx.atomic = 1
    while ctx.pc != len(code.labeled_ops):
        code.labeled_ops[ctx.pc].op.eval(state, ctx)
    legacy_harmony.constants[name] = ctx.pop()


def _parse_string(string: str, filename: str = "<string-code>") -> BlockAST:
    _input = _build_input_stream(str_value=string)
    parser = _build_parser(_input)
    visitor = HarmonyVisitorImpl(filename)

    tree = parser.program()
    return visitor.visit(tree)


def _parse(filename: str) -> BlockAST:
    _input = _build_input_stream(filename=filename)
    error_listener = HarmonyParserErrorListener(filename)
    lexer_error_listener = HarmonyLexerErrorListener(filename)
    parser = _build_parser(
        _input,
        lexer_error_listener=lexer_error_listener,
        parser_error_listener=error_listener
    )

    tree = parser.program()
    if error_listener.errors or lexer_error_listener.errors:
        raise HarmonyCompilerErrorCollection(
            lexer_error_listener.errors + error_listener.errors
        )

    visitor = HarmonyVisitorImpl(filename)
    try:
        return visitor.visit(tree)
    except HarmonyCompilerError as e:
        raise HarmonyCompilerErrorCollection([e.token])


def parse(filename: str) -> AST:
    return _parse(filename)


def parse_string(string: str) -> AST:
    return _parse_string(string)


def do_compile(fname: str, consts: List[str], mods: List[str], interface: Optional[str]):
    for c in consts:
        try:
            i = c.index("=")
            _parse_constant(c[0:i], c[i + 1:])
        except (IndexError, ValueError):
            raise HarmonyCompilerError(
                message="Usage: -c C=V to define a constant"
            )

    for m in mods:
        try:
            i = m.index("=")
            legacy_harmony.modules[m[0:i]] = m[i + 1:]
        except (IndexError, ValueError):
            raise HarmonyCompilerError(
                message="Usage: -m module=version to specify a module version"
            )

    scope = Scope(None)
    scope.file = str(fname)
    scope.inherit = True
    code = Code()
    code.modpush("__main__")
    _load_file(str(fname), scope, code, True)

    unused_constant_def = legacy_harmony.constants.keys() - legacy_harmony.used_constants
    unused_module_def = legacy_harmony.modules.keys() - legacy_harmony.imported.keys()
    if len(unused_constant_def) > 0:
        raise HarmonyCompilerError(
            message="The following constants were defined from the command line but not used: " + ', '.join(unused_constant_def),
            filename=fname,
            line=0,
            column=0,
            lexeme="",
        )
    if len(unused_module_def) > 0:
        raise HarmonyCompilerError(
            message="The following modules were defined from the command line but not used: " + ', '.join(unused_module_def),
            filename=fname,
            line=0,
            column=0,
            lexeme="",
        )

    # Analyze liveness of variables
    newcode = code.liveness()

    newcode.link()
    legacy_harmony.optimize(newcode)
    return newcode, scope



# Generated from Harmony.g4 by ANTLR 4.9.3
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .HarmonyParser import HarmonyParser
else:
    from HarmonyParser import HarmonyParser

# This class defines a complete generic visitor for a parse tree produced by HarmonyParser.

class HarmonyVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by HarmonyParser#program.
    def visitProgram(self, ctx:HarmonyParser.ProgramContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#import_stmt.
    def visitImport_stmt(self, ctx:HarmonyParser.Import_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#import_name.
    def visitImport_name(self, ctx:HarmonyParser.Import_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#import_from.
    def visitImport_from(self, ctx:HarmonyParser.Import_fromContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#import_names_seq.
    def visitImport_names_seq(self, ctx:HarmonyParser.Import_names_seqContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#tuple_bound.
    def visitTuple_bound(self, ctx:HarmonyParser.Tuple_boundContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#bound.
    def visitBound(self, ctx:HarmonyParser.BoundContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#arith_op.
    def visitArith_op(self, ctx:HarmonyParser.Arith_opContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#comp_op.
    def visitComp_op(self, ctx:HarmonyParser.Comp_opContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#unary_op.
    def visitUnary_op(self, ctx:HarmonyParser.Unary_opContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#int.
    def visitInt(self, ctx:HarmonyParser.IntContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#bool.
    def visitBool(self, ctx:HarmonyParser.BoolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#atom.
    def visitAtom(self, ctx:HarmonyParser.AtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#name.
    def visitName(self, ctx:HarmonyParser.NameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#str.
    def visitStr(self, ctx:HarmonyParser.StrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#none.
    def visitNone(self, ctx:HarmonyParser.NoneContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#set_rule_1.
    def visitSet_rule_1(self, ctx:HarmonyParser.Set_rule_1Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#empty_dict.
    def visitEmpty_dict(self, ctx:HarmonyParser.Empty_dictContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#paren_tuple.
    def visitParen_tuple(self, ctx:HarmonyParser.Paren_tupleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#bracket_tuple.
    def visitBracket_tuple(self, ctx:HarmonyParser.Bracket_tupleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#lambda_expr.
    def visitLambda_expr(self, ctx:HarmonyParser.Lambda_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#set_rule.
    def visitSet_rule(self, ctx:HarmonyParser.Set_ruleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#iter_parse.
    def visitIter_parse(self, ctx:HarmonyParser.Iter_parseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#for_parse.
    def visitFor_parse(self, ctx:HarmonyParser.For_parseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#where_parse.
    def visitWhere_parse(self, ctx:HarmonyParser.Where_parseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#tuple_rule.
    def visitTuple_rule(self, ctx:HarmonyParser.Tuple_ruleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#nary_expr.
    def visitNary_expr(self, ctx:HarmonyParser.Nary_exprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#expr_rule.
    def visitExpr_rule(self, ctx:HarmonyParser.Expr_ruleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#application.
    def visitApplication(self, ctx:HarmonyParser.ApplicationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#expr.
    def visitExpr(self, ctx:HarmonyParser.ExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#assign_op.
    def visitAssign_op(self, ctx:HarmonyParser.Assign_opContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#aug_assign_op.
    def visitAug_assign_op(self, ctx:HarmonyParser.Aug_assign_opContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#expr_stmt.
    def visitExpr_stmt(self, ctx:HarmonyParser.Expr_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#assign_stmt.
    def visitAssign_stmt(self, ctx:HarmonyParser.Assign_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#aug_assign_stmt.
    def visitAug_assign_stmt(self, ctx:HarmonyParser.Aug_assign_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#const_assign_stmt.
    def visitConst_assign_stmt(self, ctx:HarmonyParser.Const_assign_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#assert_stmt.
    def visitAssert_stmt(self, ctx:HarmonyParser.Assert_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#await_stmt.
    def visitAwait_stmt(self, ctx:HarmonyParser.Await_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#var_stmt.
    def visitVar_stmt(self, ctx:HarmonyParser.Var_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#trap_stmt.
    def visitTrap_stmt(self, ctx:HarmonyParser.Trap_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#pass_stmt.
    def visitPass_stmt(self, ctx:HarmonyParser.Pass_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#finally_stmt.
    def visitFinally_stmt(self, ctx:HarmonyParser.Finally_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#invariant_stmt.
    def visitInvariant_stmt(self, ctx:HarmonyParser.Invariant_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#del_stmt.
    def visitDel_stmt(self, ctx:HarmonyParser.Del_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#spawn_stmt.
    def visitSpawn_stmt(self, ctx:HarmonyParser.Spawn_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#go_stmt.
    def visitGo_stmt(self, ctx:HarmonyParser.Go_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#print_stmt.
    def visitPrint_stmt(self, ctx:HarmonyParser.Print_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#sequential_stmt.
    def visitSequential_stmt(self, ctx:HarmonyParser.Sequential_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#global_stmt.
    def visitGlobal_stmt(self, ctx:HarmonyParser.Global_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#builtin_stmt.
    def visitBuiltin_stmt(self, ctx:HarmonyParser.Builtin_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#atomic_block.
    def visitAtomic_block(self, ctx:HarmonyParser.Atomic_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#for_block.
    def visitFor_block(self, ctx:HarmonyParser.For_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#let_decl.
    def visitLet_decl(self, ctx:HarmonyParser.Let_declContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#when_decl.
    def visitWhen_decl(self, ctx:HarmonyParser.When_declContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#let_when_decl.
    def visitLet_when_decl(self, ctx:HarmonyParser.Let_when_declContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#let_when_block.
    def visitLet_when_block(self, ctx:HarmonyParser.Let_when_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#opt_returns.
    def visitOpt_returns(self, ctx:HarmonyParser.Opt_returnsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#method_decl.
    def visitMethod_decl(self, ctx:HarmonyParser.Method_declContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#while_block.
    def visitWhile_block(self, ctx:HarmonyParser.While_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#elif_block.
    def visitElif_block(self, ctx:HarmonyParser.Elif_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#else_block.
    def visitElse_block(self, ctx:HarmonyParser.Else_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#if_block.
    def visitIf_block(self, ctx:HarmonyParser.If_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#block_stmts.
    def visitBlock_stmts(self, ctx:HarmonyParser.Block_stmtsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#block.
    def visitBlock(self, ctx:HarmonyParser.BlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#normal_block.
    def visitNormal_block(self, ctx:HarmonyParser.Normal_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#simple_stmt.
    def visitSimple_stmt(self, ctx:HarmonyParser.Simple_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#compound_stmt.
    def visitCompound_stmt(self, ctx:HarmonyParser.Compound_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#one_line_stmt.
    def visitOne_line_stmt(self, ctx:HarmonyParser.One_line_stmtContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#label.
    def visitLabel(self, ctx:HarmonyParser.LabelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by HarmonyParser#stmt.
    def visitStmt(self, ctx:HarmonyParser.StmtContext):
        return self.visitChildren(ctx)



del HarmonyParser


import os
import os.path
from pathlib import Path
import pathlib
import sys
from typing import List, NamedTuple

import setuptools
import distutils.errors
from setuptools import Extension
from setuptools.command.build_ext import build_ext


# Checks the version number of package in the same way as pip.
def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path), encoding='utf-8') as fp:
        return fp.read()

def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")

PACKAGE_NAME = 'harmony'
PROJECT_DIR_NAME = 'harmony_model_checker'

python_version = sys.version_info[:2]
if python_version < (3, 6):
    print("{} requires Python version 3.6 or later".format(PACKAGE_NAME))
    print("(Version {}.{} detected)".format(*python_version))
    sys.exit(1)

PACKAGE_VERSION = get_version(f"{PROJECT_DIR_NAME}/__init__.py")

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


class CompilerArgs(NamedTuple):
    name: str
    compile_args: List[str]
    linker_args: List[str]


EXTRA_COMPILE_ARGS = ["-pthread", "-m64", "-O3", "-DNDEBUG", "-fPIC"] # the -fPIC flag is required to create a shared object file.

# Extra link args used by gcc in Linux.
# Tested flags with cc and clang
EXTRA_LINK_ARGS = ['-pthread', '-shared', '-Wl,-O1', '-Wl,-Bsymbolic-functions', '-Wl,-Bsymbolic-functions', '-Wl,-z,relro', '-g', '-fwrapv', '-O2', '-Wl,-Bsymbolic-functions',
                   '-Wl,-z,relro', '-g', '-fwrapv', '-O2', '-g', '-fstack-protector-strong', '-Wformat', '-Werror=format-security', '-Wdate-time', '-D_FORTIFY_SOURCE=2']

compiler_and_args = [
    CompilerArgs(
        "gcc",
        EXTRA_COMPILE_ARGS,
        EXTRA_LINK_ARGS
    ),
    CompilerArgs(
        "cc",
        EXTRA_COMPILE_ARGS,
        EXTRA_LINK_ARGS
    ),
    CompilerArgs(
        "clang",
        EXTRA_COMPILE_ARGS,
        EXTRA_LINK_ARGS
    )
]

class BuildExtCommand(build_ext):
    def build_extension(self, ext) -> None:
        try:
            # Try to build the extension with the default OS build tools.
            super().build_extension(ext)
            return
        except (distutils.errors.DistutilsPlatformError, distutils.errors.CompileError) as e:
            print(e)
            print("Encountered error when building by default configurations")
            print("Building with backup configurations")

        encountered_error = None
        for c in compiler_and_args:
            ext.extra_compile_args = c.compile_args
            ext.extra_link_args = c.linker_args
            name = c.name
            self.compiler.set_executable("linker_so", name)
            self.compiler.set_executable("compiler_so", name)
            self.compiler.set_executable("compiler_cxx", name)
            try:
                super().build_extension(ext)
                return
            except distutils.errors.CompileError as e:
                encountered_error = e

        if encountered_error is not None:
            raise encountered_error


CHARM_PKG_PATH = pathlib.Path(PROJECT_DIR_NAME) / "charm"
def get_c_extension_src():
    sources = []
    base_dir = CHARM_PKG_PATH
    D = [base_dir, base_dir / "python_ext"]
    for d in D:
        sources.extend([str(f) for f in d.glob('*.c')])
    return sources

def get_c_extension_include_dirs():
    return [
        str(CHARM_PKG_PATH),
        str(CHARM_PKG_PATH / 'python_ext'),
    ]


module = Extension(
    f"{PROJECT_DIR_NAME}.charm",
    sources=get_c_extension_src(),
    include_dirs=get_c_extension_include_dirs(),
    extra_compile_args=[],
    extra_link_args=[]
    # extra_link_args=
    #     ["-lnuma"] if sys.platform.lower().startswith("linux") else []
)

setuptools.setup(
    name=PACKAGE_NAME,
    version=PACKAGE_VERSION,
    author="Robbert van Renesse",
    author_email="rvr@cs.cornell.edu",
    description="Harmony Programming Language",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    install_requires=[
        'antlr-denter>=1.3.1',
        'antlr4-python3-runtime==4.9.3',
        'automata-lib',
        'pydot',
        'requests',
    ],
    license="BSD",
    url="https://harmony.cs.cornell.edu",
    include_package_data=True,
    package_data={
        PACKAGE_NAME: [
            "modules/*.hny",
            "charm/*.h",
            "charm/*.c",
            "charm/python_ext/ext.c",
            "harmony/charm.js",
            "harmony/charm.css",
        ]
    },
    entry_points={
        'console_scripts': [
            'harmony=harmony_model_checker.main:main'
        ]
    },
    python_requires=">=3.6",
    ext_modules=[module],
    cmdclass={
        "build_ext": BuildExtCommand
    }
)



import sys
import json
from typing import Dict, Optional
from automata.fa.dfa import DFA # type: ignore

from harmony_model_checker.iface import Transitions_t

def parse(file):
    with open(file, encoding='utf-8') as f:
        js = json.load(f)

        states = { "__error__": '' }
        initial_state: Optional[str] = None
        final_states = set()
        input_symbols = set()
        transitions: Dict[str, Dict[str, str]] = { "__error__": {} }

        for s in js["nodes"]:
            idx = str(s["idx"])
            val = str(s["value"])
            input_symbols.add(val)
            transitions[idx] = {}
            if s["type"] == "initial":
                assert initial_state is None
                initial_state = idx
            elif s["type"] == "terminal":
                final_states.add(idx)
            states[idx] = val

        for s in states:
            for x in input_symbols:
                transitions[s][x] = "__error__"

        for edge in js['edges']:
            src = str(edge["src"])
            dst = str(edge["dst"])
            assert dst != initial_state
            assert src not in final_states
            if src != dst:
                if dst in final_states:
                    val = "__term__"
                else:
                    val = states[dst]
                transitions[src][val] = dst

    return DFA(
        states=set(states.keys()),
        input_symbols=input_symbols,
        transitions=transitions,
        initial_state=initial_state,
        final_states=final_states
    )

def main():
    print("DFA1")
    dfa1 = parse(sys.argv[1])
    print("DFA2")
    dfa2 = parse(sys.argv[2])
    print("CMP")
    assert dfa1 <= dfa2
    print("DONE")

if __name__ == "__main__":
    main()



from typing import Set
from harmony_model_checker.harmony.ops import *

class Labeled_Op:
    def __init__(self, module: Optional[str], op: Op, start, stop, stmt, labels):
        self.module = module    # module
        self.op = op            # operation
        self.start = start      # first token
        self.stop = stop        # last token
        self.stmt = stmt
        self.labels = labels
        self.live_in: Set[str] = set()
        self.live_out: Set[str] = set()

class Code:
    def __init__(self, parent=None):
        self.labeled_ops: List[Labeled_Op] = []
        self.endlabels = set()
        self.modstack = []      # module stack
        self.curModule = None
        self.curFile = None
        self.curLine = 0
        self.parent = parent

    def modpush(self, module):
        self.modstack.append(self.curModule)
        self.curModule = module

    def modpop(self):
        self.curModule = self.modstack.pop()

    def location(self, file, line: int):
        self.curFile = file
        self.curLine = line

    def append(self, op: Op, start, stop, labels=set(), stmt=None):
        assert len(start) == 4
        assert len(stop) == 4
        assert self.curModule is not None
        self.labeled_ops.append(Labeled_Op(self.curModule, op, start, stop, stmt, labels | self.endlabels))
        self.endlabels = set()

    def nextLabel(self, endlabel):
        self.endlabels.add(endlabel)

    # This method inserts DelVar operations as soon as a variable is no
    # longer live
    def liveness(self):
        # First figure out what the labels point to and initialize
        # the nodes
        map = {}
        lop_predecessors: Dict[int, Set[int]] = {}
        for pc in range(len(self.labeled_ops)):
            lop = self.labeled_ops[pc]
            lop_predecessors[pc] = set()
            lop.live_in = set()
            lop.live_out = set()
            for label in lop.labels:
                assert label not in map, label
                map[label] = pc
        # Compute the predecessors of each node
        for pc in range(len(self.labeled_ops)):
            lop = self.labeled_ops[pc]
            if isinstance(lop.op, JumpOp):
                assert isinstance(lop.op.pc, LabelValue)
                op_pc = map[lop.op.pc]
                succ = self.labeled_ops[op_pc]
                lop_predecessors[op_pc] |= {pc}
            elif isinstance(lop.op, JumpCondOp):
                assert pc < len(self.labeled_ops) - 1
                assert isinstance(lop.op.pc, LabelValue)
                op_pc = map[lop.op.pc]
                succ = self.labeled_ops[op_pc]
                lop_predecessors[op_pc] |= {pc}
                lop_predecessors[pc + 1] |= {pc}
            elif pc < len(self.labeled_ops) - 1 and not isinstance(lop.op, ReturnOp):
                lop_predecessors[pc + 1] |= {pc}
        # Live variable analysis
        change = True
        while change:
            change = False
            for pc in range(len(self.labeled_ops)):
                lop = self.labeled_ops[pc]
                if pc == len(self.labeled_ops) - 1:
                    live_out = set()
                elif isinstance(lop.op, JumpOp):
                    assert isinstance(lop.op.pc, LabelValue)
                    succ = self.labeled_ops[map[lop.op.pc]]
                    live_out = succ.live_in
                else:
                    live_out = self.labeled_ops[pc + 1].live_in
                    if isinstance(lop.op, JumpCondOp):
                        assert isinstance(lop.op.pc, LabelValue)
                        succ = self.labeled_ops[map[lop.op.pc]]
                        live_out = live_out | succ.live_in
                live_in = lop.op.use() | (live_out - lop.op.define())
                if not change and (live_in != lop.live_in or live_out != lop.live_out):
                    change = True
                lop.live_in = live_in
                lop.live_out = live_out
        # Create new code with DelVars inserted
        newcode = Code()
        for lop_pc, lop in enumerate(self.labeled_ops):
            # print(lop.op, lop.live_in, lop.live_out)

            # If a variable is live on output of any predecessor but not
            # live on input, delete it first
            pre_del = set()
            for pred in lop_predecessors[lop_pc]:
                plop = self.labeled_ops[pred]
                live_out = plop.live_out | plop.op.define()
                pre_del |= live_out - lop.live_in

            labels = lop.labels
            newcode.curModule = lop.module

            # If an atomic inc operation, put the delvar inside the atomic section
            if isinstance(lop.op, AtomicIncOp):
                newcode.append(lop.op, lop.start, lop.stop, labels=labels, stmt=lop.stmt)
                labels = set()

            for d in sorted(pre_del - { 'this' }):
                newcode.append(DelVarOp((d, None, None, None)), lop.start, lop.stop, labels=labels, stmt=lop.stmt)
                labels = set()

            if not isinstance(lop.op, AtomicIncOp):
                newcode.append(lop.op, lop.start, lop.stop, labels=labels, stmt=lop.stmt)

            # If a variable is defined or live on input but not live on output,
            # immediately delete afterward
            # TODO.  Can optimize StoreVar by replacing it with Pop
            # post_del = (lop.op.define() | lop.live_in) - lop.live_out
            post_del = lop.live_in - lop.live_out
            for d in sorted(post_del - { 'this' }):
                newcode.append(DelVarOp((d, None, None, None)), lop.start, lop.stop, stmt=lop.stmt)

        return newcode

    def link(self):
        map = {}
        for pc in range(len(self.labeled_ops)):
            lop = self.labeled_ops[pc]
            for label in lop.labels:
                assert label not in map, label
                map[label] = PcValue(pc)
        for lop in self.labeled_ops:
            lop.op.substitute(map)



from abc import abstractmethod
import json
from typing import Any, List, Optional, Tuple

tlavarcnt = 0           # to generate unique TLA+ variables

def tlaValue(lexeme):
    if isinstance(lexeme, bool):
        return 'HBool(%s)'%("TRUE" if lexeme else "FALSE")
    if isinstance(lexeme, int):
        return 'HInt(%d)'%lexeme
    if isinstance(lexeme, str):
        return 'HStr("%s")'%lexeme
    return lexeme.tlaval()

def strValue(v):
    if isinstance(v, Value) or isinstance(v, bool) or isinstance(v, int) or isinstance(v, float):
        return str(v)
    if isinstance(v, str):
        return '"%s"'%v
    assert False, v

def jsonValue(v):
    if isinstance(v, Value):
        return v.jdump()
    if isinstance(v, bool):
        return '{ "type": "bool", "value": "%s" }'%str(v)
    if isinstance(v, int) or isinstance(v, float):
        return '{ "type": "int", "value": %s }'%str(v)
    if isinstance(v, str):
        return '{ "type": "atom", "value": %s }'%json.dumps(v, ensure_ascii=False)
    assert False, v

def strVars(v):
    assert isinstance(v, DictValue)
    result = ""
    for (var, val) in v.d.items():
        if result != "":
            result += ", "
        result += strValue(var)[1:] + "=" + strValue(val)
    return "{" + result + "}"

def keyValue(v):
    if isinstance(v, bool):
        return (0, v)
    if isinstance(v, int):
        return (1, v)
    if isinstance(v, str):
        return (2, v)
    assert isinstance(v, Value), v
    return v.key()

def substValue(v, map):
    return v.substitute(map) if isinstance(v, Value) else v

class Value:
    def __str__(self):
        return self.__repr__()

    def jdump(self):
        assert False

    def substitute(self, map):
        assert False, self
    
    @abstractmethod
    def key(self) -> Tuple[int, int]:
        pass

class PcValue(Value):
    def __init__(self, pc):
        assert isinstance(pc, int), pc
        self.pc = pc

    def __repr__(self):
        return "PC(" + str(self.pc) + ")"

    def tlaval(self):
        return 'HPc(%d)'%self.pc

    def __hash__(self):
        return self.pc.__hash__()

    def __eq__(self, other):
        return isinstance(other, PcValue) and other.pc == self.pc

    def key(self):
        return (3, self.pc)

    def jdump(self):
        return '{ "type": "pc", "value": "%d" }'%self.pc

    def substitute(self, map):
        # return substValue(self.pc, map)
        return self

class _LabelIdGenerator:

    _next_id = 0

    @classmethod
    def new_id(cls):
        result = cls._next_id
        cls._next_id += 1
        return result

# This is a substitute for PCValues used before values are assigned to labels
# TODO.  Get rid of all but id
class LabelValue(Value):

    def __init__(self, module, label):
        self.id = _LabelIdGenerator.new_id()
        self.module = module
        self.label = label

    def __repr__(self):
        if self.module is None:
            return "LABEL(" + str(self.id) + ", " + self.label + ")"
        else:
            return "LABEL(" + self.module + ":" + self.label + ")"

    def __hash__(self):
        return self.id

    def __eq__(self, other):
        return isinstance(other, LabelValue) and other.id == self.id

    def key(self):
        return (100, self.id)

    def jdump(self):
        assert False, self

    def substitute(self, map):
        return map[self] if self in map else self

class ListValue(Value):
    def __init__(self, vals):
        self.vals = vals

    def __repr__(self):
        if self.vals == []:
            return "()"
        result = ""
        for v in self.vals:
            if result != "":
                result += ", "
            result += strValue(v)
        if len(self.vals) == 1:
            result += ","
        return "[" + result + "]"

    def tlaval(self):
        s = "<<" + ",".join(tlaValue(x) for x in self.vals) + ">>"
        return 'HList(%s)'%s

    def jdump(self):
        result = ""
        for v in self.vals:
            if result != "":
                result += ", "
            result += jsonValue(v)
        return '{ "type": "list", "value": [%s] }'%result

    def __hash__(self):
        hash = 0
        for v in self.vals:
            hash ^= v.__hash__()
        return hash

    def __eq__(self, other):
        if not isinstance(other, ListValue):
            return False
        return self.vals == other.vals

    def __len__(self):
        return len(self.vals)

    def key(self):
        return (4, [ keyValue(v) for v in self.vals ])

    def substitute(self, map):
        return ListValue([ substValue(v, map) for v in self.vals ])

class DictValue(Value):
    def __init__(self, d: dict):
        self.d = d

    def __repr__(self):
        if len(self.d) == 0:
            return "{:}"
        result = ""
        if set(self.d.keys()) == set(range(len(self.d))):
            for k in range(len(self.d)):
                if result != "":
                    result += ", "
                result += strValue(self.d[k])
            return "[" + result + "]"
        keys = sorted(self.d.keys(), key=keyValue)
        for k in keys:
            if result != "":
                result += ", "
            result += strValue(k) + ":" + strValue(self.d[k])
        return "{ " + result + " }"

    def tlaval(self):
        global tlavarcnt

        tlavar = "x%d"%tlavarcnt
        tlavarcnt += 1
        if len(self.d) == 0:
            return 'EmptyDict'
        s = "[ %s \\in {"%tlavar + ",".join({ tlaValue(k) for k in self.d.keys() }) + "} |-> "
        # special case: all values are the same
        vals = list(self.d.values())
        if vals.count(vals[0]) == len(vals):
            s += tlaValue(vals[0]) + " ]"
            return 'HDict(%s)'%s

        # not all values are the same
        first = True
        for k,v in self.d.items():
            if first:
                s += "CASE "
                first = False
            else:
                s += " [] "
            s += "%s = "%tlavar + tlaValue(k) + " -> " + tlaValue(v)
        s += " [] OTHER -> FALSE ]"
        return 'HDict(%s)'%s

    def jdump(self):
        result = ""
        keys = sorted(self.d.keys(), key=keyValue)
        for k in keys:
            if result != "":
                result += ", "
            result += '{ "key": %s, "value": %s }'%(jsonValue(k), jsonValue(self.d[k]))
        return '{ "type": "dict", "value": [%s] }'%result

    def __hash__(self):
        hash = 0
        for x in self.d.items():
            hash ^= x.__hash__()
        return hash

    def __eq__(self, other):
        if not isinstance(other, DictValue):
            return False
        return self.d == other.d

    def __len__(self):
        return len(self.d)

    # Dictionary ordering generalizes lexicographical ordering when the dictionary
    # represents a list or tuple
    def key(self):
        return (5, [ (keyValue(v), keyValue(self.d[v]))
                        for v in sorted(self.d.keys(), key=keyValue)])

    def substitute(self, map):
        return DictValue({ substValue(k, map): substValue(v, map)
                            for (k, v) in self.d.items() })

emptytuple = ListValue([])
emptydict = DictValue({})

class SetValue(Value):
    def __init__(self, s):
        self.s = s

    def __repr__(self):
        if len(self.s) == 0:
            return "{}"
        result = ""
        vals = sorted(self.s, key=keyValue)
        for v in vals:
            if result != "":
                result += ", "
            result += strValue(v)
        return "{ " + result + " }"

    def tlaval(self):
        s = "{" + ",".join(tlaValue(x) for x in self.s) + "}"
        return 'HSet(%s)'%s

    def jdump(self):
        result = ""
        vals = sorted(self.s, key=keyValue)
        for v in vals:
            if result != "":
                result += ", "
            result += jsonValue(v)
        return '{ "type": "set", "value": [%s] }'%result

    def __hash__(self):
        return frozenset(self.s).__hash__()

    def __eq__(self, other):
        if not isinstance(other, SetValue):
            return False
        return self.s == other.s

    def key(self):
        return (6, [keyValue(v) for v in sorted(self.s, key=keyValue)])

    def substitute(self, map):
        return SetValue({ substValue(v, map) for v in self.s })

# An address is a closure consisting of a function and a list of arguments
# All are of type Value.  Due to backwards compatibility, we also have to
# represent "None" as an AddressValue.  For this we use func=None and args=[].
class AddressValue(Value):
    def __init__(self, func, args):
        self.func = func
        self.args = args

    def remainder(self):
        result = ""
        for index in self.args[1:]:
            result += "[" + strValue(index) + "]"
        return result

    def indexes(self):
        return [self.func] + self.args

    def __repr__(self):
        if self.func is None:
            assert self.args == []
            return "None"
        # assert self.args != []
        if isinstance(self.func, PcValue):
            if self.func.pc in { -1, -2 }:      # shared or method variable
                return "?" + self.args[0] + self.remainder()
            if self.func.pc == -3:              # thread-local variable
                return "?this." + self.args[0] + self.remainder()
        result = "?" + strValue(self.func)
        for index in self.args:
            result += "[" + strValue(index) + "]"
        return result

    def tlaval(self):
        if self.func is None:
            return "None"
        s = "<<" + ",".join(tlaValue(x) for x in self.args) + ">>"
        return 'Address(%s, %s)'%(tlaValue(self.func), s)

    def jdump(self):
        if self.func is None:
            return '{ "type": "address" }'
        result = ""
        for index in self.args:
            if result != "":
                result += ", "
            result = result + jsonValue(index)
        return '{ "type": "address", "func": %s, "args": [%s] }'%(jsonValue(self.func), result)

    def __hash__(self):
        hash = 0
        for x in self.args:
            hash ^= x.__hash__()
        return hash

    def __eq__(self, other):
        if not isinstance(other, AddressValue):
            return False
        return self.args == other.args

    def key(self):
        return (7, self.args)

    def substitute(self, map):
        return AddressValue(substValue(self.func, map), [ substValue(v, map) for v in self.args ])

class ContextValue(Value):
    def __init__(self, name: Tuple[str, Optional[str], Optional[int], Optional[int]], entry: int, arg, this):
        self.name = name
        self.entry = entry
        self.arg = arg
        self.pc = entry
        self.this = this
        self.atomic = 0
        self.readonly = 0
        self.interruptLevel = False
        self.stack: List[Any] = []     # collections.deque() seems slightly slower
        self.fp = 0         # frame pointer
        self.vars = emptydict
        self.trap = None
        self.phase = "start"        # start, middle, or end
        self.stopped = False
        self.failure: Optional[str] = None

    def __repr__(self):
        return "ContextValue(" + str(self.name) + ", " + str(self.arg) + ", " + str(self.this) + ")"

    def __str__(self):
        return self.__repr__()

    def nametag(self):
        if self.arg == emptytuple:
            return self.name[0] + "()"
        return self.name[0] + "(" + str(self.arg) + ")"

    def __hash__(self):
        h = (self.name, self.entry, self.arg, self.pc, self.this, self.atomic, self.readonly, self.interruptLevel, self.vars,
            self.trap, self.phase, self.stopped, self.failure).__hash__()
        for v in self.stack:
            h ^= v.__hash__()
        return h

    def __eq__(self, other):
        if not isinstance(other, ContextValue):
            return False
        if self.name != other.name:
            return False
        if self.entry != other.entry:
            return False
        if self.arg != other.arg:
            return False
        if self.pc != other.pc:
            return False
        if self.this != other.this:
            return False
        if self.atomic != other.atomic:
            return False
        if self.readonly != other.readonly:
            return False
        if self.interruptLevel != other.interruptLevel:
            return False
        if self.phase != other.phase:
            return False
        if self.stopped != other.stopped:
            return False
        if self.fp != other.fp:
            return False
        if self.trap != other.trap:
            return False
        if self.failure != other.failure:
            return False
        return self.stack == other.stack and self.vars == other.vars

    def copy(self):
        c = ContextValue(self.name, self.entry, self.arg, self.this)
        c.pc = self.pc
        c.atomic = self.atomic
        c.readonly = self.readonly
        c.interruptLevel = self.interruptLevel
        c.stack = self.stack.copy()
        c.fp = self.fp
        c.trap = self.trap
        c.vars = self.vars
        c.phase = self.phase
        c.stopped = self.stopped
        c.failure = self.failure
        return c

    def get(self, var: str):
        return self.this if var == "this" else self.vars.d[var]

    def iget(self, indexes):
        assert indexes[0] != "this"
        v = self.vars
        while indexes != []:
            v = v.d[indexes[0]]
            indexes = indexes[1:]
        return v

    def update(self, record, indexes, val):
        if len(indexes) > 1:
            v = self.update(record.d[indexes[0]], indexes[1:], val)
        else:
            v = val
        d = record.d.copy()
        d[indexes[0]] = v
        return DictValue(d)

    def doDelete(self, record, indexes):
        if len(indexes) > 1:
            d = record.d.copy()
            d[indexes[0]] = self.doDelete(record.d[indexes[0]], indexes[1:])
        else:
            d = record.d.copy()
            if indexes[0] in d:
                del d[indexes[0]]
        return DictValue(d)

    def set(self, indexes, val):
        if indexes[0] == "this":
            if len(indexes) == 1:
                self.this = val
            else:
                self.this = self.update(self.this, indexes[1:], val)
        else:
            self.vars = self.update(self.vars, indexes, val)

    def delete(self, indexes):
        self.vars = self.doDelete(self.vars, indexes)

    def push(self, val):
        assert val is not None
        self.stack.append(val)

    def pop(self):
        return self.stack.pop()

    def key(self):
        return (8, (self.pc, self.__hash__()))



from typing import List
from antlr4.error.ErrorListener import ErrorListener  # type: ignore

from harmony_model_checker.exception import ErrorToken
from harmony_model_checker.parser.HarmonyParser import HarmonyParser

class HarmonyLexerErrorListener(ErrorListener):
    def __init__(self, filename: str):
        super().__init__()
        self.filename = filename
        self.errors: List[ErrorToken] = []

    def syntaxError(self, recognizer, offending_symbol, line, column, msg, e):
        lexeme = str(offending_symbol.text) if offending_symbol and hasattr(offending_symbol, 'text') else ""
        self.errors.append(ErrorToken(
            filename=self.filename,
            lexeme=lexeme,
            message=str(msg) + " syntax error",
            line=line,
            column=column + 1,
            is_eof_error=False
        ))

class HarmonyParserErrorListener(ErrorListener):
    def __init__(self, filename: str):
        super().__init__()
        self.filename = filename
        self.errors: List[ErrorToken] = []

    def syntaxError(self, recognizer, offending_symbol, line, column, msg, e):
        lexeme = offending_symbol.text
        line = line
        column = column + 1

        if offending_symbol.type == HarmonyParser.INDENT:
            self.errors.append(ErrorToken(
                filename=self.filename,
                lexeme=lexeme,
                message="Indentation error",
                line=line,
                column=column,
                is_eof_error=False
            ))
            return
        
        if offending_symbol.type == HarmonyParser.DEDENT:
            self.errors.append(ErrorToken(
                filename=self.filename,
                lexeme=lexeme,
                message="Unexpected dedent. May be caused by indentation error.",
                line=line,
                column=column,
                is_eof_error=False
            ))
            return

        if isinstance(msg, str):
            if msg.startswith("no viable alternative at input"):
                msg = f"Unexpected token {lexeme}"
            elif msg.startswith("extraneous input"):
                msg = f"Extraneous input {lexeme}. May be caused by another error."
            elif msg.startswith("mismatched input"):
                msg = f"Unexpected token {lexeme}"

        self.errors.append(ErrorToken(
            filename=self.filename,
            lexeme=lexeme,
            message=str(msg),
            line=line,
            column=column,
            is_eof_error=False
        ))



import pathlib
import json

from harmony_model_checker.harmony.behavior import behavior_parse
# from harmony_model_checker.harmony.summary import summaryMain
from harmony_model_checker.harmony.summarize import Summarize

def brief_kv(js):
    return (brief_string(js["key"]), brief_string(js["value"]))

def brief_idx(js):
    return "[" + brief_string(js) + "]"

def brief_string(js):
    type = js["type"]
    if type == "address":
        if "func" not in js:
            return "None"
        result = "?"
        func = js["func"]
        args = js["args"]
        if func["type"] == "pc":
            if int(func["value"]) == -1:
                result += args[0]["value"]
                args = args[1:]
            elif int(func["value"]) == -2:
                result += "@" + args[0]["value"]
                args = args[1:]
            else:
                result += brief_string(func)
        else:
            result += brief_string(func)
        return result + "".join([ brief_idx(kv) for kv in args ])
    v = js["value"]
    if type == "bool":
        return v
    if type == "int":
        return str(v) if isinstance(v, int) else v
    if type == "atom":
        return json.dumps(v, ensure_ascii=False)
    if type == "set":
        if v == []:
            return "{}"
        lst = [ brief_string(val) for val in v ]
        return "{ " + ", ".join(lst) + " }"
    if type == "list":
        if v == []:
            return "[]"
        lst = [ brief_string(val) for val in v ]
        return "[ " + ", ".join(lst) + " ]"
    if type == "dict":
        if v == []:
            return "{:}"
        lst = [ brief_kv(kv) for kv in v ]
        keys = [ k for k,v in lst ]
        if keys == [str(i) for i in range(len(v))]:
            return "[ " + ", ".join([v for k,v in lst]) + " ]" 
        else:
            return "{ " + ", ".join([k + ": " + v for k,v in lst]) + " }" 
    if type == "pc":
        return "PC(%s)"%v
    if type == "context":
        return "CONTEXT(" + str(v["pc"]) + ")"

def brief_print_vars(d):
    print("{", end="")
    first = True
    for k, v in d.items():
        if first:
            first = False
        else:
            print(",", end="")
        print(" %s: %s"%(k, brief_string(v)), end="")
    print(" }")

def brief_print_range(start, end):
    if start == end:
        return "%d"%(start)
    if start + 1 == end:
        return "%d,%d"%(start, end)
    return "%d-%d"%(start, end)

class Brief:
    def __init__(self):
        self.tid = None
        self.name = None
        self.start = 0
        self.steps = ""
        self.interrupted = False
        self.lastmis = {}
        self.shared = {}
        self.failure = ""

    def flush(self):
        if self.tid is not None:
            print("T%s: %s ["%(self.tid, self.name), end="")
            if self.steps != "":
                self.steps += ","
            self.steps += brief_print_range(self.start, int(self.lastmis["pc"]))
            print(self.steps + "] ", end="")
            brief_print_vars(self.shared)

    def print_macrostep(self, mas):
        mis = mas["microsteps"]
        if mas["tid"] != self.tid:
            self.flush()
            self.tid = mas["tid"]
            self.name = mas["name"]
            self.interrupted = False
            self.lastmis = mis[0]
            self.start = int(self.lastmis["pc"])
            if "shared" in self.lastmis:
                self.shared = self.lastmis["shared"]
            lastpc = 0
            self.steps = ""
            begin = 1
        else:
            begin = 0
        for i in range(begin, len(mis)):
            if "shared" in mis[i]:
                self.shared = mis[i]["shared"]
            if self.interrupted:
                if self.steps != "":
                    self.steps += ","
                self.steps += brief_print_range(self.start, int(self.lastmis["pc"]))
                self.start = int(mis[i]["pc"])
                self.steps += ",interrupt"
            elif "choose" in mis[i]:
                if self.steps != "":
                    self.steps += ","
                self.steps += brief_print_range(self.start, int(mis[i]["pc"]))
                self.steps += "(choose %s)"%brief_string(mis[i]["choose"])
                self.start = int(mis[i]["pc"]) + 1
            elif "print" in mis[i]:
                if self.steps != "":
                    self.steps += ","
                self.steps += brief_print_range(self.start, int(mis[i]["pc"]))
                self.steps += "(print %s)"%brief_string(mis[i]["print"])
                self.start = int(mis[i]["pc"]) + 1
            elif int(mis[i]["pc"]) != int(self.lastmis["pc"]) + 1:
                if self.steps != "":
                    self.steps += ","
                self.steps += brief_print_range(self.start, int(self.lastmis["pc"]))
                self.start = int(mis[i]["pc"])
            self.lastmis = mis[i]
            if "failure" in self.lastmis:
                self.failure = self.lastmis["failure"]
            self.interrupted = "interrupt" in self.lastmis and self.lastmis["interrupt"] == "True"

    def run(self, outputfiles, behavior):
        with open(outputfiles["hco"], encoding='utf-8') as f:
            print("* Phase 6: loading", outputfiles["hco"])
            top = json.load(f, strict=False)
            assert isinstance(top, dict)
            if top["issue"] == "No issues":
                behavior_parse(top, False, outputfiles, behavior)
            else:
                se = Summarize()
                se.run(outputfiles, top)

            # print()
            # p = pathlib.Path(outputfiles["htm"]).resolve()
            # url = "file://" + str(p)
            # print("open " + url + " for detailed information, or use the HarmonyGUI")

            # print("Issue:", top["issue"])
            # assert isinstance(top["macrosteps"], list)
            # for mes in top["macrosteps"]:
            #     self.print_macrostep(mes)
            # self.flush()
            # print(self.failure)
            # print("* Phase 7: print failure summary")
            # print(summaryMain(outputfiles, top))
